# Copyright Louis Paternault 2015
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Compilation using commands depending on mimetype or file extensions"""

import mimetypes
import operator
import pathlib
import shlex
import re

from evariste.plugins.action.command import Command, CommandCompiler

class AutoCmdCompiler(CommandCompiler):
    """Compilation using rules based on extension and mimetypes"""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._setup = self.parent.get_command(self.path.from_fs)['setup']

    def get_target(self):
        return (
            self.path.parent.from_source.as_posix()
            /
            pathlib.Path(self.path.format(self._setup['target']))
            )

    def iter_commands(self):
        yield from sorted([self._setup[key] for key in self._setup if key.startswith('command')])

class AutoCmd(Command):
    """Compilation using rules based on mimetypes and extensions"""

    keyword = "action.autocmd"
    priority = 50
    pathcompiler = AutoCmdCompiler

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        # Parsing setup to build the command dictionary
        self._commands = {}
        for (key, setup) in self.shared.setup.items():
            if key.startswith(self.keyword):
                local = {
                    'setup': setup
                    }

                if setup['priority'] is None:
                    local['priority'] = (1, key)
                else:
                    local['priority'] = (0, int(setup['priority']))
                local['mimetypes'] = setup['mimetypes']
                if local['mimetypes'] is None:
                    local['mimetypes'] = ""
                local['extensions'] = setup['extensions']
                if local['extensions'] is None:
                    local['extensions'] = ""

                local['mimetypes'] = [
                    re.compile(regexp)
                    for regexp
                    in shlex.split(local['mimetypes'])
                    ]
                local['extensions'] = [
                    "." + ext
                    for ext
                    in shlex.split(local['extensions'])
                    ]
                if (not local['mimetypes']) and (not local['extensions']):
                    local['extensions'] = [key[len(self.keyword):]]
                self._commands[key] = local

    def get_command(self, value):
        """Return the command associated with the given file.

        May return ``None`` if no command is associated with it.
        """
        for setup in sorted(self._commands.values(), key=operator.itemgetter('priority')):
            if value.suffix in setup['extensions']:
                return setup
            for regexp in setup['mimetypes']:
                mime = mimetypes.guess_type(str(value))[0]
                if isinstance(mime, str):
                    if regexp.match(mime):
                        return setup
        return None

    def match(self, value):
        return self.get_command(value) is not None



#!/bin/env python3

# Copyright Louis Paternault 2015
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Command line client to :mod:`evs`"""

import os

import argdispatch

import evariste

def _execlp(program, args):
    """Call :func:`os.execlp`, adding `program` as the first argument to itself."""
    return os.execlp(program, program, *args)

def commandline_parser():
    """Return a command line parser."""

    parser = argdispatch.ArgumentParser(
        prog="evs",
        description=(
            "Helper script for the evariste tool."
            ),
        formatter_class=argdispatch.RawTextHelpFormatter,
        )

    parser.add_argument(
        '--version',
        action='version',
        help='Show version',
        version='%(prog)s ' + evariste.VERSION
        )

    subparsers = parser.add_subparsers(
        title="Subcommands",
        description=(
            "List of available subcommands, either built-in, or gathered from `evs-*` binaries."
            ),
        )
    subparsers.required = True
    subparsers.dest = "subcommand"
    subparsers.add_submodules("evariste.evs")
    subparsers.add_prefix_executables("evs-")

    return parser

def main():
    """Main function"""

    commandline_parser().parse_args()

if __name__ == "__main__":
    main()



# Copyright Louis Paternault 2015
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Dummy, do-nothing vcs. Used for tests."""

from datetime import datetime
import os

from evariste.plugins.vcs import VCS

class NoneVCS(VCS):
    """Dummy vcs"""

    keyword = "vcs.none"

    def walk(self):
        yield from []

    def is_versionned(self, path):
        return False

    @property
    def workdir(self):
        return os.getcwd()

    def from_repo(self, path):
        return os.path.relpath(
            path.resolve().as_posix(),
            self.workdir
            )

    def last_modified(self, path):
        return datetime.fromtimestamp(os.path.getmtime(str(path)))



# Copyright Louis Paternault 2015
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Render (processed) tree: abstract class."""

import glob
import os
import pathlib
import tempfile
import textwrap

from evariste import errors, plugins
from evariste.plugins import NoMatch

################################################################################
# Targets: how should target files be rendererd?
class TargetRenderer(plugins.PluginBase):
    """Renderer of target (compiled) files."""

    def match(self, dummy):
        """This is the default renderer, that matches everything."""
        return True

    def render(self, tree):
        """Render ``tree``, which is a :class:`tree.File`."""
        raise NotImplementedError()

################################################################################
# Readme plugins: annotation of files

class CannotRender(errors.EvaristeError):
    """Exception raised when renderer cannot render a given file."""
    pass

class ReadmeRenderer(plugins.PluginBase):
    """Generic class of README renderers, and also a default raw renderer."""
    # pylint: disable=too-few-public-methods, abstract-method

    extensions = []

    def render(self, path, *args, **kwargs):
        """Render the readme file ``path``.
        """
        # pylint: disable=unused-argument, no-self-use
        with open(path) as file:
            return file.read()

    def iter_readme(self, value, directory):
        """Iterave over potential readme files for the path ``value``."""
        if directory:
            yield from self.iter_readme_dir(value)
        else:
            yield from self.iter_readme_file(value)

    def iter_readme_dir(self, value):
        """Iterate over potential readme for the given directory ``value``."""
        for ext in self.extensions:
            for filename in glob.iglob((value / "*.{}".format(ext)).as_posix()):
                basename = os.path.basename(filename)
                if basename.count(".") == 1:
                    if basename.split(".")[0].lower() == "readme":
                        yield filename

    def iter_readme_file(self, value):
        """Iterate over potential readme for the given file ``value``."""
        for ext in self.extensions:
            for filename in glob.iglob("{}.{}".format(value, ext)):
                yield filename

################################################################################
# Renderer: renders the directory tree into something (html, etc.)

class Renderer(plugins.PluginBase):
    """Abstract tree renderer."""

    plugin_type = "renderer"

    subplugins = {
        "readme": ReadmeRenderer,
        "target": TargetRenderer,
        }

    global_default_setup = {'renderer': {'enable_plugins': ''}}

    def priority(self):
        """Return keyword as priority.

        That way, plugins are sorted by name.
        """
        return self.keyword

    def render(self, tree):
        """Render tree."""
        raise NotImplementedError()

    def iter_subplugins(self, subtype):
        """Iterate over subplugins of type `subtype`."""
        for parent in self.__class__.mro(): # pylint: disable=no-member
            if not hasattr(parent, 'keyword'):
                break
            try:
                yield from self.shared.builder.plugins.get_plugin_type(
                    "{}.{}".format(parent.keyword, subtype)
                    ).values()
            except NoMatch:
                pass

    def prune_readmes(self, tree):
        """Remove readme files from tree."""
        for filename in list(tree.iter_readmes(self.iter_subplugins('readme'))):
            if tree.find(filename):
                tree.prune(filename)



# Copyright Louis Paternault 2015
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Share global data between evariste objects."""

from enum import Enum

from evariste.utils import DeepDict

class ViewPoint(Enum):
    """Point of view: from plugin or from tree?"""
    # pylint: disable=too-few-public-methods
    plugin = 1
    tree = 2

ATTRIBUTES = {
    'setup': [ViewPoint.plugin],
    'plugin': [ViewPoint.plugin],
    'tree': [ViewPoint.tree, ViewPoint.plugin],
    }

class SharedDeepDict(DeepDict):
    """DeepDict class, with a factory to build them as shared attributes."""

    @classmethod
    def from_attr(cls, attr, dictionary=None):
        """Default value for attribute."""
        return cls(len(ATTRIBUTES[attr]), dictionary)

class _SubSharedView:
    """View of  a shared data.

    :param str keyword: Keyword of the point of view (e.g. 'renderer.html').
    :param str attr: Object being viewed (e.g. 'setup').
    :param ViewPoint viewpoint: Point of view (tree, or plugin).
    :param Shared shared: Shared data.
        """
    # pylint: disable=too-few-public-methods

    def __init__(self, keyword, attr, viewpoint, shared):
        self.keyword = keyword
        self.viewpoint = viewpoint
        self.shared = shared
        self.attr = attr


class _SubSharedViewNoDict(_SubSharedView):
    """View of a shared data, not as a dictionary."""
    # pylint: disable=too-few-public-methods

    def __getitem__(self, key):
        return getattr(self.shared, self.attr)[key]

    def __setitem__(self, key, value):
        getattr(self.shared, self.attr)[key] = value

class _SubSharedViewDict(_SubSharedView):
    """View of a shared data, as a dictionary (of dictionaries of ...)"""
    # pylint: disable=too-few-public-methods

    def __getitem__(self, key):
        shared = getattr(self.shared, self.attr)
        for viewpoint in ATTRIBUTES[self.attr]:
            if viewpoint == self.viewpoint:
                shared = shared[self.keyword]
            else:
                shared = shared[key]
        return shared

    def __setitem__(self, key, value):
        shared = getattr(self.shared, self.attr)
        for viewpoint in ATTRIBUTES[self.attr][:-1]:
            if viewpoint == self.viewpoint:
                shared = shared[self.keyword]
            else:
                shared = shared[key]
        if ATTRIBUTES[self.attr][-1] == self.viewpoint:
            shared[self.keyword] = value
        else:
            shared[key] = value

class _MetaSharedView(type):
    """Metaclass that creates attributes corresponding to :data:`ATTRIBUTES`.
    """
    def __new__(mcs, *args, **kwargs):
        obj = super().__new__(mcs, *args, **kwargs)
        for attr in ATTRIBUTES:
            setattr(obj, attr, property(
                fget=obj.getter(attr),
                fset=obj.setter(attr),
                ))

        return obj

class _SharedView(metaclass=_MetaSharedView):
    """View of shared data, from a certain point of view."""
    # pylint: disable=too-few-public-methods

    def __init__(self, viewpoint, keyword, shared):
        self.shared = shared
        self.keyword = keyword
        self.viewpoint = viewpoint
        self.views = {}
        for attr in ATTRIBUTES:
            if viewpoint in ATTRIBUTES[attr]:
                if len(ATTRIBUTES[attr]) == 1:
                    self.views[attr] = _SubSharedViewNoDict(keyword, attr, viewpoint, shared)
                else:
                    self.views[attr] = _SubSharedViewDict(keyword, attr, viewpoint, shared)

    @classmethod
    def getter(cls, attr):
        """Returns a getter for this attribute of this point of view."""
        def getter(self):
            """Getter"""
            if isinstance(self.views[attr], _SubSharedViewNoDict):
                return self.views[attr][self.keyword]
            else:
                return self.views[attr]
        return getter

    @classmethod
    def setter(cls, attr):
        """Returns a setter for this attribute of this point of view."""
        def setter(self, value):
            """Setter"""
            if isinstance(self.views[attr], _SubSharedViewNoDict):
                self.views[attr][self.keyword] = value
            else:
                # Should not be that difficult to implement, but I do not need
                # it right now.
                raise NotImplementedError
        return setter

class Shared:
    """Shared data"""

    def __init__(self, builder, **kwargs):
        self.builder = builder
        for attr in ATTRIBUTES:
            if attr in kwargs:
                setattr(self, attr, kwargs[attr])
            else:
                setattr(self, attr, SharedDeepDict.from_attr(attr))

    def get_plugin_view(self, keyword):
        """Get this data, from the point of view of a plugin."""
        return _SharedView(ViewPoint.plugin, keyword, self)

    def get_tree_view(self, path):
        """Get this data, from the point of view of a tree."""
        return _SharedView(ViewPoint.tree, path, self)



# Copyright Louis Paternault 2017
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""IkiWiki markdown README renderer"""

import markdown

from evariste.plugins.renderer.html.readme import HtmlRenderer

class MdwnRenderer(HtmlRenderer):
    """Markdown renderer for readme files."""

    keyword = "renderer.html.readme.mdwn"
    extensions = ["mdwn", "md"]

    def render(self, path, context):
        """Render file as html code.
        """
        with open(path, encoding="utf8") as source:
            return markdown.markdown(source.read())



# Copyright Louis Paternault 2015-2017
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Parse setup file."""

import collections
import configparser

from evariste import utils
from evariste.errors import EvaristeError

def Options(*args, **kwargs):
    """Return a :class:`collections.defaultdict` object, default being `None`.
    """
    # pylint: disable=invalid-name
    return collections.defaultdict(lambda: None, *args, **kwargs)

def Sections(dictionary=None):
    """Return a :class:`collections.defaultdict` object, default being `Options`
    """
    # pylint: disable=invalid-name
    if dictionary is None:
        dictionary = {}
    return collections.defaultdict(
        Options,
        {key: Options(value) for (key, value) in dictionary.items()}
        )

class Setup:
    """Representation ef Evariste setup."""

    def __init__(self, dictionary=None):
        if dictionary is None:
            dictionary = {}
        self.dict = Sections(dictionary)

    def __iter__(self):
        yield from self.dict

    def __getitem__(self, value):
        return self.dict[value]

    def __str__(self):
        return "{{{}}}".format(", ".join([
            "{}: {}".format(key, value)
            for key, value
            in self.dict.items()
            ]))

    def __eq__(self, other):
        for section in set(self.keys()) | set(other.keys()):
            for option in set(self[section].keys()) | set(other[section].keys()):
                if self[section][option] != other[section][option]:
                    return False
        return True

    def fill_blanks(self, dictionary):
        """Fill unset self options with argument.

        Returns ``self`` for convenience.
        """
        for section in dictionary:
            if section in self.dict:
                for option in dictionary[section]:
                    if option not in self.dict[section]:
                        self.dict[section][option] = dictionary[section][option]
            else:
                self.dict[section] = Options(dictionary[section])
        return self

    def update(self, other, *, extend_list=False):
        """Update setup, with another setup or dict object.

        Similar to :meth:`dict.update`.

        :param boolean extend_list: If `True` and both values are lists,
            extend the first one with the second one.

        Return `self` for convenience.
        """
        for section in other:
            for option in other[section]:
                if (
                        extend_list
                        and
                        isinstance(self.dict[section][option], list)
                        and
                        isinstance(other[section][option], list)
                    ):
                    self.dict[section][option].extend(other[section][option])
                else:
                    self.dict[section][option] = other[section][option]
        return self


    @classmethod
    def from_file(cls, filename):
        """Parse configuration file ``filename``."""
        with open(filename, encoding='utf8') as file:
            return cls.from_string(file.read())

    @classmethod
    def from_string(cls, string):
        """Parse ``string`` as the content of a configuration file."""
        config = configparser.ConfigParser(allow_no_value=True)
        config.read_string(string)
        return cls.from_config(config)

    @classmethod
    def from_config(cls, setup):
        """Parse ``setup`` as a :class:`configparser.ConfigParser` object."""
        return cls.from_dict(utils.DeepDict.from_configparser(setup))

    @classmethod
    def from_dict(cls, dictionary):
        """Return a setup object, from a dictionary."""
        return cls(dictionary)

    def keys(self):
        """Return a new view of the setup’s keys"""
        return self.dict.keys()

    def pprint(self):
        """Pretty print of the object."""
        from pprint import pprint
        pprint(self.dict)

    @property
    def __dict__(self):
        """Return self, as a :class:`dict` of :class:`dict`."""
        dictionary = {}
        for section in self:
            dictionary[section] = dict(self[section])
        return dictionary

    def copy(self):
        """Return a copy of this object."""
        return self.__class__.from_dict(vars(self))

    def items(self):
        """Iterate over (key, value) pairs, as 2-tuples."""
        yield from self.dict.items()


class SetupError(EvaristeError):
    """Error in setup file."""

    pass



# Copyright Louis Paternault 2015
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Plugin to copy files at the end of compilation and rendering."""

import glob

from evariste import plugins, utils
from evariste.hooks import ContextHook

class _Close(ContextHook):
    """Hook to copy files at the end of compilation."""

    def exit(self, value=None):
        for key in self.local.setup:
            if key.startswith("copy"):
                if isinstance(self.local.setup[key], str):
                    arguments = self.local.setup[key].split()
                else:
                    arguments = self.local.setup[key]
                try:
                    sources, dest = arguments
                except (ValueError, TypeError):
                    raise ValueError(
                        """[copy] Option "{}" should be "SOURCE DEST".""".format(key)
                        )
                if dest is None:
                    continue
                for path in glob.iglob(sources):
                    utils.copytree(path, dest)
        return value

class Copy(plugins.PluginBase):
    """Copy files at the end of compilation."""
    # pylint: disable=too-few-public-methods

    plugin_type = ""
    keyword = "copy"
    hooks = {
        'close': _Close,
        }



# Copyright Louis Paternault 2016
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Access to any file on the file system."""

import os

from evariste.plugins.vcs import VCS

class FS(VCS):
    """Access all files on the file system."""

    keyword = "vcs.fs"

    def walk(self, *, root=None):
        if root is None:
            root = self.workdir
        for dirpath, dirnames, filenames in os.walk(root):
            for dirname in dirnames:
                yield from self.walk(root=os.path.join(root, dirname))
            for filename in filenames:
                yield os.path.relpath(
                    os.path.join(dirpath, filename),
                    self.workdir,
                    )

    def is_versionned(self, path):
        return path.is_file() and (self.workdir in path.parents)

    @property
    def workdir(self):
        return str(self.source)



# Copyright Louis Paternault 2015
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Test of the ``libdirs`` setup option"""

import os

from evariste.plugins import PluginNotFound

from . import TestLoadedPlugins

def absdir(path):
    """Return the absolute path of the argument (relative to this file)."""
    return os.path.join(
        os.path.dirname(__file__),
        "libdirs",
        path,
        )

class TestLibdirs(TestLoadedPlugins):
    """Testing setup `libdirs` options"""
    # pylint: disable=too-few-public-methods

    def test_nolibdirs(self):
        """Test compilation without the libdirs option"""
        self.assertSetEqual(
            self._loaded_plugins({}),
            self.default_plugins,
            )

    def test_fail(self):
        """Test failure (plugin not found)."""
        with self.assertRaises(PluginNotFound):
            self._loaded_plugins({
                'setup': {
                    'libdirs': absdir("plugins"),
                    'enable_plugins': ["foo", "bar"],
                    },
                })

    def test_libdirs(self):
        """Test compilation with the libdirs option"""
        with self.subTest("Without spaces"):
            self.assertSetEqual(
                self._loaded_plugins({
                    'setup': {
                        'libdirs': absdir("plugins"),
                        'enable_plugins': ['foo'],
                        },
                    }),
                set([
                    'foo',
                    ]) | self.default_plugins
                )

        with self.subTest("With spaces"):
            self.assertSetEqual(
                self._loaded_plugins({
                    'setup': {
                        'libdirs': '"{}"'.format(absdir("plug ins")),
                        'enable_plugins': ['bar'],
                        },
                    }),
                set([
                    "bar",
                    ]) | self.default_plugins
                )

        with self.subTest("With and without spaces"):
            self.assertSetEqual(
                self._loaded_plugins({
                    'setup': {
                        'libdirs': '{} {}'.format(
                            absdir(r"plug\ ins"),
                            absdir("plugins"),
                            ),
                        'enable_plugins': ["foo", "bar"],
                        },
                    }),
                set([
                    "bar",
                    "foo",
                    ]) | self.default_plugins
                )



# Copyright Louis Paternault 2015
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Render tree as an HTML (body) page."""

import logging

from evariste import VERSION
from evariste.plugins.renderer.html import target, readme
from evariste.plugins.renderer.jinja2 import JinjaRenderer
from evariste.utils import expand_path, smart_open

class HTMLRenderer(JinjaRenderer):
    """Render tree as an HTML div (without the `<div>` tags)."""
    # pylint: disable=too-few-public-methods

    keyword = "renderer.html"
    extension = "html"
    description = "Render as an html div"
    default_setup = {
        "href_prefix": "",
        "destfile": 'index.html',
        }
    depends = {
        'suggested': [
            'renderer.html.readme.html',
            'renderer.html.target.default',
        ],
    }
    default_templatevar = {
        "aftertree": 'Generated using <a href="http://framagit.org/spalax/evariste">Évariste</a> version {}.'.format(VERSION), # pylint: disable=line-too-long
    }

    def render(self, tree):
        rendered = super().render(tree)
        with smart_open(expand_path(self.local.setup['destfile']), 'w') as destfile:
            destfile.write(rendered)
        return rendered



# Copyright Louis Paternault 2015
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Test involving compilation"""

import collections
import difflib
import filecmp
import os
import re
import sys
import unittest

from evariste.builder import Builder
from evariste.setup import Setup

class Compare:
    """Compare some characteristics of files."""
    # pylint: disable=too-few-public-methods

    message = ""

    @staticmethod
    def compare(first, second):
        """Compare files. Return `True` iff they differ."""
        raise NotImplementedError()

class FirstOnly(Compare):
    """File is present in left directory only."""
    # pylint: disable=abstract-method, too-few-public-methods

    message = "Only in first"

class SecondOnly(Compare):
    """File is present in right directory only."""
    # pylint: disable=abstract-method, too-few-public-methods

    message = "Only in second"

class SameContent(Compare):
    """Files have same content"""
    # pylint: disable=too-few-public-methods

    message = "File content differ"

    @staticmethod
    def compare(left, right):
        result = filecmp.cmp(left, right)
        if not result:
            with open(left) as leftfile:
                with open(right) as rightfile:
                    sys.stderr.writelines(difflib.Differ().compare(
                        leftfile.readlines(),
                        rightfile.readlines(),
                        ))
        return result


class TestCompilation(unittest.TestCase):
    """Test of compilation"""

    @staticmethod
    def _run_evariste(local):
        """Create a builder, from a dummy setup file.

        Return a set a loaded plugins.
        """
        setup = Setup.from_dict({
            'setup': {
                'vcs': 'vcs.git',
                },
            'vcs': {
                'enable_plugins': ['vcs.git'],
                },
            })
        setup.update(local, extend_list=True)

        with Builder.from_setupdict(setup) as builder:
            builder.compile()
            for renderer in builder.iter_renderers():
                renderer()

    def assertTreeEqual(self, first, second, rules=None):
        """Compare if directory trees `first` and `second` are equal.

        :param list rules: List of tuples `(regexp, Compare)` objects, for
            special rules concerning objects. When comparing a file, the first
            tuple with a regexp matching the file (base) name is considered,
            and its associated `Compare` class is used to compare files. By
            default, the :class:`SameContent` class is used. This is useful to
            define custom matches (for instance, two generated PDF will not be
            exactly equal, so a fuzzy comparison may be used).
        """
        # pylint: disable=invalid-name
        if rules is None:
            rules = list()
        rules.append((r".*", SameContent))
        rules = [
            (re.compile(regexp), comparator)
            for regexp, comparator
            in rules
            ]
        errors = self.compare_dirs(first, second, rules)
        if sum([len(files) for files in errors.values()]) != 0:
            raise AssertionError(
                "\n".join([
                    "{message}:\n\t{files}".format(
                        message=error.message,
                        files="\n\t".join(errors[error])
                        )
                    for error in errors
                    ])
                )

    def compare_dirs(self, first, second, rules):
        """Compare directories, and return a dictionary of differences."""
        errors = collections.defaultdict(list)

        first_set = set(os.listdir(first))
        second_set = set(os.listdir(second))
        for filename in first_set - second_set:
            errors[FirstOnly].append(filename)
        for filename in second_set - first_set:
            errors[SecondOnly].append(filename)

        for path in first_set & second_set:
            if os.path.isdir(os.path.join(first, path)):
                for (errorname, filenames) in self.compare_dirs(
                        os.path.join(first, path),
                        os.path.join(second, path),
                        rules).items():
                    for name in filenames:
                        errors[errorname].append(os.path.join(path, name))
            else:
                for regexp, comparator in rules:
                    if regexp.match(path):
                        if not comparator.compare(
                                os.path.join(first, path),
                                os.path.join(second, path),
                            ):
                            errors[comparator].append(path)
                        break

        return errors



# Copyright Louis Paternault 2015
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Access to git-versionned files."""

from datetime import datetime
import logging
import os

import git

from evariste.plugins.vcs import VCS, NoRepositoryError

LOGGER = logging.getLogger(__name__)

class Git(VCS):
    """Access git-versionned files"""
    # pylint: disable=no-member

    keyword = "vcs.git"

    def __init__(self, shared):
        super().__init__(shared)
        try:
            self.repository = git.Repo(str(self.source), search_parent_directories=True)
        except git.InvalidGitRepositoryError:
            raise NoRepositoryError(self.keyword, self.source)

        self._last_modified = self._read_modification_date()

    def _read_modification_date(self):
        """Return a dictionary of versionned files and their last modification time."""
        LOGGER.info("Reading git commit dates…")
        last_modified = {}
        no_known_date = set((blob.path for blob in self.repository.tree().traverse()))
        for commit in self.repository.iter_commits():
            new = no_known_date & commit.stats.files.keys()
            for path in new:
                last_modified[path] = datetime.fromtimestamp(commit.committed_date)
                no_known_date.remove(path)
            if not no_known_date:
                break
        for blob in self.repository.index.iter_blobs():
            if blob[1].path not in last_modified:
                last_modified[blob[1].path] = super().last_modified(
                    os.path.join(self.workdir, blob[1].path)
                    )
        LOGGER.info("Done")
        return last_modified


    def walk(self):
        source = self.source.resolve().as_posix()
        for entry in self._last_modified:
            path = os.path.join(self.workdir, entry)
            if path.startswith(source):
                yield os.path.relpath(path, source)

    def is_versionned(self, path):
        return self.from_repo(path) in self._last_modified

    @property
    def workdir(self):
        return self.repository.working_dir

    def last_modified(self, path):
        if not self.is_versionned(path):
            return super().last_modified(path)

        return self._last_modified[self.from_repo(path)]



# Copyright Louis Paternault 2015
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Set of plugins testing plugin dependency, and enabling/disabling plugins.

Does nothing, but printing stuff to standard output.
"""

from datetime import datetime
import os
import pathlib

from evariste import plugins

class TestPlugin(plugins.PluginBase):
    """Generic test plugin"""

    plugin_type = ""

################################################################################
# Dummy plugins

class Foo(TestPlugin):
    keyword = "foo"

class Bar(TestPlugin):
    keyword = "bar"

################################################################################
# Default / Required plugins

class Required(TestPlugin):
    keyword = "required"
    required = True

class Default(TestPlugin):
    keyword = "default"
    default = True

################################################################################
# Dependencies : First simple plugins

class DependsRequired(TestPlugin):
    keyword = "dependsrequired"
    depends = {
        'required': [
            'foo',
        ],
    }

class DependsOptional(TestPlugin):
    keyword = "dependsoptional"
    depends = {
        'suggested': [
            'foo',
        ],
    }

class DependsBoth(TestPlugin):
    keyword = "dependsboth"
    depends = {
        'required': [
            'foo',
        ],
        'suggested': [
            'bar',
        ],
    }

################################################################################
# Dependecies : Recursive dependency

class DependsOptionalOptional(TestPlugin):
    keyword = "dependsoptionaloptional"
    depends = {
        'suggested': [
            'dependsoptional',
        ],
    }

class DependsOptionalRequired(TestPlugin):
    keyword = "dependsoptionalrequired"
    depends = {
        'suggested': [
            'dependsrequired',
        ],
    }

class DependsRequiredRequired(TestPlugin):
    keyword = "dependsrequiredrequired"
    depends = {
        'required': [
            'dependsrequired',
        ],
    }

class DependsRequiredOptional(TestPlugin):
    keyword = "dependsrequiredoptional"
    depends = {
        'required': [
            'dependsoptional',
        ],
    }



# Copyright Louis Paternault 2015
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Abstract utilities for target renderers using Jinja2."""

from evariste.plugins.renderer import TargetRenderer

class Jinja2TargetRenderer(TargetRenderer):
    """Renderer of target (compiled) files using jinja2"""

    keyword = None
    extension = None
    priority = -float("inf")

    def match(self, dummy):
        """This is the default renderer, that matches everything."""
        return True

    def render(self, tree, context):
        """Render ``tree``, which is a :class:`tree.File`."""
        # pylint: disable=arguments-differ
        return context['render'](
            context,
            tree,
            ["tree_file.{}".format(self.extension)],
            )



# Copyright Louis Paternault 2015
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Rest README renderer"""

from docutils import core

from evariste.plugins.renderer.html.readme import HtmlRenderer

class RestRenderer(HtmlRenderer):
    """reStructuredText renderer for readme files."""
    # pylint: disable=too-few-public-methods, abstract-method

    keyword = "renderer.html.readme.rst"
    extensions = ["rst", "rest"]

    def render(self, path, context):
        """Render file as html code.
        """
        with open(path) as source:
            return core.publish_parts(
                source=source.read(),
                source_path=path,
                #destination_path=destination_path,
                writer_name='html',
                settings_overrides={
                    #'input_encoding': input_encoding,
                    'doctitle_xform': False,
                    'initial_header_level': 1,
                    }
                )['body']



# Copyright Louis Paternault 2017
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Actions performed to compile files."""

import io
import logging
import os

from evariste import errors, plugins
from evariste.hooks import ContextHook

LOGGER = logging.getLogger(__name__)

################################################################################
# Cache log

class CacheLogHook(ContextHook):
    """Hook called before and after compilation of one file."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.tree = None

    def enter(self, tree, *_args, **_kwargs):
        self.tree = tree

    def exit(self):
        """Store the log in cache."""
        tree = self.tree

        # Saving log into cache
        self.local.tree[tree.from_source] = tree.report.log.getvalue()

class CacheLog(plugins.PluginBase):
    """Plugin to cache compilation log."""
    # pylint: disable=too-few-public-methods

    plugin_type = ''
    keyword = "cachelog"
    hooks = {
        'compilefile': CacheLogHook,
        }

################################################################################
# Actions

class PathCompiler:
    """Compiler of one path.

    The main difference between this object and :class:`Action` is that:
        - there is one instance of :class:`Action` for the whole tree;
        - there is one instance of :class:`PathCompiler` per path.
    """
    # pylint: disable=too-few-public-methods

    def __init__(self, parent, path, keyword, shared):
        self.shared = shared
        self.keyword = keyword
        self.path = path
        self.parent = parent

    def compile(self):
        """Perform actual compilation of `self.path`."""
        raise NotImplementedError()

class Action(plugins.PluginBase):
    """Generic action"""
    # pylint: disable=too-many-instance-attributes, too-few-public-methods

    plugin_type = "action"
    pathcompiler = PathCompiler

    def compile(self, path):
        """Compile ``path``, catching :class:`EvaristeError` exceptions.

        This function *must* be thread-safe.
        """
        try:
            return self.pathcompiler(self, path, self.keyword, self.shared).compile()
        except errors.EvaristeError as error:
            LOGGER.error(str(error))
            return Report(
                path,
                success=False,
                log=io.StringIO(str(error)),
                )

class DirectoryCompiler(PathCompiler):
    """Compiler of a directory."""

    def compile(self):
        if self.success:
            log = ""
        else:
            log = "At least one file in this directory failed."
        return Report(
            self.path,
            success=self.success(),
            log=log,
            target=None,
            )

    def success(self):
        """Return ``True`` if compilation of all subpath succeeded."""
        for sub in self.path:
            if not self.path[sub].report.success:
                return False
        return True


class DirectoryAction(Action):
    """Fake action on directories."""
    # pylint: disable=abstract-method, too-few-public-methods

    keyword = "action.directory"
    pathcompiler = DirectoryCompiler
    required = True

    def match(self, dummy):
        return False

################################################################################
# Reports

class Report:
    """Report of an action. Mainly a namespace with very few methods."""

    def __init__(
            self,
            path,
            target=None,
            success=False,
            log=None,
            depends=None,
        ):
        # pylint: disable=too-many-arguments

        self.depends = depends
        if self.depends is None:
            self.depends = set()

        self.log = log
        if self.log is None:
            self.log = io.StringIO()

        self.path = path
        self.target = target
        self._success = success

    @property
    def full_depends(self):
        """Set of files this action depends on, including ``self.path``."""
        return self.depends | set([self.path.from_fs])

    @property
    def success(self):
        """Success getter"""
        return self._success

    @success.setter
    def success(self, value):
        """Success setter."""
        self._success = value



# Copyright Louis Paternault 2015
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Set of classes extending the standard :mod:`threading` module."""

import contextlib
import functools
import queue
import threading

class Lock:
    """A lock, with contexts :meth:`thread_safe` and :meth:`thread_unsafe`.
    """

    def __init__(self):
        self._lock = threading.RLock()

    @contextlib.contextmanager
    def thread_unsafe(self):
        """Context marking the inner context as thread unsafe.

        This means that no other thread sharing this lock is run in parallel.
        """
        self._lock.acquire()
        yield
        self._lock.release()

    @contextlib.contextmanager
    def thread_safe(self):
        """Context marking the inner context as thread safe.

        This means that other threads sharing this lock can run in parallel.
        """
        self._lock.release()
        yield
        self._lock.acquire()

@functools.total_ordering
class Task:
    """A task to run, with a priority."""
    # pylint: disable=too-few-public-methods

    def __init__(self, priority, *functions):
        self.priority = priority
        self.functions = functions

    def run(self):
        """Run the (list of) tasks."""
        for function in self.functions:
            function()

    def __eq__(self, other):
        return self.priority == other.priority

    def __lt__(self, other):
        return self.priority < other.priority

    def __str__(self):
        return str(self.functions)

class Pool:
    """A pool of threads.

    Can be used as a context, to create and to start threads when entering the
    context, and waiting for the threads to finish when exiting the context.

    Tasks (as :class:`Task`) may be put into the :attr:`Pool.queue`. They will
    be executed by the different threads running in parallel.
    """
    # pylint: disable=too-few-public-methods

    def __init__(self, workers):
        self.queue = Queue(maxsize=2*workers)
        self._threads = list()
        for __ignored in range(workers):
            self._threads.append(
                threading.Thread(
                    target=self.worker,
                    daemon=True,
                    )
                )

    def __enter__(self):
        for thread in self._threads:
            thread.start()
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        for __ignored in self._threads:
            # Sending "end" signal to threads
            self.queue.put(None)
        self.queue.join()
        for thread in self._threads:
            thread.join()
        return False

    def worker(self):
        """Function called in a thread, running tasks got from the queue."""
        while True:
            with self.queue.get() as task:
                if task is None:
                    return
                else:
                    task()

class Queue(queue.Queue):
    """A subclass of :class:`queue.Queue`, with :meth:`get` as a context."""

    @contextlib.contextmanager
    def get(self):
        """Like :meth:`queue.Queue.get`, but calling :meth:`task_done` at the end of the context."""
        # pylint: disable=arguments-differ
        item = super().get()
        yield item
        self.task_done()



# Copyright Louis Paternault 2016
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


"""Text renderer: print report on terminal."""

import logging
import sys

import unidecode

from evariste.plugins.renderer import Renderer
from evariste.utils import yesno
import evariste

LOGGER = logging.getLogger(evariste.__name__)

class TreePrettyPrinter:
    """Pretty print a tree"""
    # pylint: disable=too-few-public-methods

    # This class could be a function, but it is a recursive function, and
    # turning it into a class allows to tell appart arguments that change
    # for every recursive call, and arguments that never change.

    def __init__(
            self,
            *,
            print_function=print,
            report=False,
            color=False,
            reverse=False,
            display=None
        ):
        self.print_function = print_function
        self.report = report
        self.color = color
        self.reverse = reverse
        self.display = display

    def pprint(self, tree, *, prefix="", last=True):
        """Pretty print the given tree"""
        # pylint: disable=too-many-branches
        line = []

        txttree = ""
        txttree += prefix
        if last:
            txttree += {
                False: "└",
                True: "┌",
                }[self.reverse]
        else:
            txttree += "├"
        txttree += "─"
        if tree.is_dir():
            txttree += {
                True: "┴",
                False: "┬",
                }[self.reverse]
        else:
            txttree += "─"
        txttree += "╼"
        line.append(txttree)

        if self.report:
            if tree.report.success:
                if self.color:
                    reportstr = "\033[92m ok \033[0m"
                else:
                    reportstr = " ok "
            else:
                if self.color and tree.is_dir():
                    reportstr = "\033[93mfail\033[0m"
                elif self.color and not tree.is_dir():
                    reportstr = "\033[91mfail\033[0m"
                else:
                    reportstr = "fail"
            reportstr = "[{}]".format(reportstr)
            line.append(reportstr)

        line.append(str(tree.relativename))

        if not self.reverse:
            self.print_function(" ".join(line))

        if last:
            postfix = "  "
        else:
            postfix = "│ "

        subtree = [branch for branch in tree.values() if self.display(branch)]
        if self.reverse:
            lastcounter = 1
        else:
            lastcounter = len(subtree)
        counter = 0
        for sub in sorted(subtree, reverse=self.reverse):
            counter += 1
            self.pprint(
                sub,
                prefix=prefix + postfix,
                last=(counter == lastcounter),
                )

        if self.reverse:
            self.print_function(" ".join(line))

ALLOWED_DISPLAY_OPTIONS = ['errors', 'all', 'errors_or_all']

class TextRenderer(Renderer):
    """Text renderer: print report on terminal."""

    keyword = 'renderer.text'
    default_setup = {
        'color': "auto",
        'ascii': False,
        'reverse': False,
        'display': 'all',
        }

    def render(self, tree):

        # Option 'ascii'
        if yesno(self.local.setup['ascii']):
            def asciiprint(text):
                """Convert argument to ascii, and print it."""
                print(unidecode.unidecode(text))
            print_function = asciiprint
        else:
            print_function = print

        # Option 'reverse'
        reverse = yesno(self.local.setup['reverse'])

        # Option 'display'
        displayoption = self.local.setup['display']
        if displayoption not in ALLOWED_DISPLAY_OPTIONS:
            LOGGER.warning(
                (
                    "[renderer.text] Option 'display' must be one of {} (currently {})."
                ).format(
                    ", ".join(("'{}'".format(option) for option in ALLOWED_DISPLAY_OPTIONS)),
                    "'{}'".format(displayoption),
                ))
            displayoption = self.default_setup['display']
        if displayoption == 'errors_or_all':
            if tree.report.success:
                displayoption = 'all'
            else:
                displayoption = 'errors'
        if displayoption == 'errors':
            display = lambda tree: not tree.report.success
        else:
            display = lambda tree: True


        # Option 'color'
        color = self.local.setup['color']
        if color.lower() == 'auto':
            color = sys.stdout.isatty() # pylint: disable=no-member
        color = yesno(color)

        # Render
        TreePrettyPrinter(
            report=True,
            color=color,
            print_function=print_function,
            reverse=reverse,
            display=display,
            ).pprint(tree)



# Copyright Louis Paternault 2015
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Compilation of {,La,Lua,Xe}TeX files"""

import pathlib

from evariste import utils
from evariste.plugins.action.command import Command, CommandCompiler

class LatexCompiler(CommandCompiler):
    """Compile path as a *latex file."""

    DEFAULT_COMMANDS = [
        "latex",
        "?bibtex",
        "?makeindex",
        "?latex",
        ]
    PREDEFINED_COMMANDS = {
        'latex': '{latex} {basename}',
        'pdflatex': 'pdflatex {basename}',
        'lualatex': 'lualatex {basename}',
        'xelatex': 'xelatex {basename}',
        'bibtex': '{bibtex} {basename}',
        'biblatex': 'biblatex {basename}',
        'makeindex': 'makeindex {basename}',
        }
    TARGET = {
        'latex': '{basename}.dvi',
        'pdflatex': '{basename}.pdf',
        'lualatex': '{basename}.pdf',
        'xelatex': '{basename}.pdf',
        }
    CONDITIONAL_COMMANDS = [
        'latex',
        'bibtex',
        'makeindex',
        ]
    BIN = {
        'latex': 'latex',
        'bibtex': 'bibtex',
        'makeindex': 'makeindex',
        }


    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.latex_todo = 1
        # Reading binaries
        self.bin = self.BIN.copy()
        for key in self.bin:
            if self.path.config[self.keyword][key] is not None:
                if self.path.config[self.keyword][key] is not None:
                    self.bin[key] = self.path.config[self.keyword][key]

    def get_target(self):
        # Setting target
        if not self.path.config[self.keyword]['target'] is not None:
            basetarget = self.path.format(self.TARGET[self.bin['latex']])
        else:
            basetarget = self.path.format(self.path.config[self.keyword]["target"])
        return self.path.parent.from_source / pathlib.Path(basetarget)

    def _format_command(self, command):
        """Iterate over commands that are to be executed."""
        if command.startswith("?") and command[1:] in self.CONDITIONAL_COMMANDS:
            yield from getattr(self, "_conditionnal_command_{}".format(command[1:]))()
        elif command.startswith("!"):
            yield utils.partial_format(command[1:], **self.bin)
        elif command in self.PREDEFINED_COMMANDS:
            yield utils.partial_format(self.PREDEFINED_COMMANDS[command], **self.bin)
        else:
            yield utils.partial_format(command, **self.bin)

    def iter_commands(self):

        precommands = []
        postcommands = []
        commands = []
        if self.keyword in self.path.config:
            for option in self.path.config[self.keyword]:
                if option.startswith("pre"):
                    precommands.append("!{}".format(self.path.config[self.keyword][option]))
                elif option.startswith("cmd"):
                    commands.append(self.path.config[self.keyword][option])
                elif option.startswith("post"):
                    postcommands.append("!{}".format(self.path.config[self.keyword][option]))
        if not commands:
            commands = self.DEFAULT_COMMANDS

        for command in precommands + commands + postcommands:
            yield from self._format_command(command)

    def _conditionnal_command_bibtex(self):
        """Yield a bibtex command, if necessary.

        "Necessary" means that option 'bibtex' appears in configuration file.

        It might be detected, some day, by looking at the LaTeX log.
        """
        if self.path.config[self.keyword]['bibtex'] is not None:
            self.latex_todo += 1
            yield from self._format_command('bibtex')

    def _conditionnal_command_makeindex(self):
        """Yield a makeindex command, if necessary.

        "Necessary" means that option 'makeindex' appears in configuration file.

        It might be detected, some day, by looking at the LaTeX log.
        """
        if self.path.config[self.keyword]['makeindex'] is not None:
            self.latex_todo += 1
            yield from self._format_command('makeindex')

    def _conditionnal_command_latex(self):
        """Yield as many `latex` commands as necessary.

        Previous commands may have incremented `self.latex_todo`.
        """
        while self.latex_todo != 0:
            self.latex_todo -= 1
            yield from self._format_command('latex')

class Latex(Command):
    """Compilation of *TeX files."""

    keyword = "action.latex"
    priority = 100
    pathcompiler = LatexCompiler

    def match(self, value):
        return value.suffix == ".tex"



# Copyright Louis Paternault 2015
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Compilation using `make` binary"""

import pathlib

from evariste.plugins.action.command import Command, CommandCompiler

class MakeCmdCompiler(CommandCompiler):
    """Compilation using `make` binary"""

    def iter_commands(self):
        yield "{bin} {options} {target}".format(
            bin=self.parent.local.setup['bin'],
            options=self.parent.local.setup['options'],
            target=pathlib.Path(self.path.format(self.path.config[self.keyword]["target"])),
            )

class MakeCmd(Command):
    """Compilation using `make` binary"""

    keyword = "action.make"
    priority = 50
    pathcompiler = MakeCmdCompiler
    default_setup = {
        'bin': 'make',
        'options': '',
        }



# Copyright Louis Paternault 2015
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Dummy plugin, used to set default command line arguments.

As evariste relies on some arguments having a value, useful when embedding
evariste into another program.
"""
import os

from evariste import plugins

class Arguments(plugins.PluginBase):
    """Set needed default command line arguments."""
    # pylint: disable=too-few-public-methods

    plugin_type = ""
    keyword = "arguments"
    global_default_setup = {
        'arguments': {
            'jobs':
            os.cpu_count()+1
        }
    }



# Copyright Louis Paternault 2015
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


"""Plugin loader"""

import collections
import contextlib
import functools
import logging
import os
import shlex
import sys

from evariste import errors, utils, setup
from evariste.hooks import set_methodhook
import evariste

LOGGER = logging.getLogger(evariste.__name__)

class NoMatch(errors.EvaristeError):
    """No plugin found matching ``value``."""

    def __init__(self, value, available):
        super().__init__()
        self.value = value
        self.available = available

    def __str__(self):
        return "Value '{}' does not match any of {}.".format(
            self.value,
            str(self.available),
            )

class SameKeyword(errors.EvaristeError):
    """Two plugins have the same keyword."""

    def __init__(self, keyword, plugin1, plugin2):
        super().__init__()
        self.keyword = keyword
        self.plugins = (plugin1, plugin2)

    def __str__(self):
        return """Plugins '{}' and '{}' have the same keyword '{}'.""".format(
            self.plugins[0].__name__,
            self.plugins[1].__name__,
            self.keyword,
            )

class NotAPlugin(errors.EvaristeError):
    """Superclass of plugins is not a plugin."""

    def __init__(self, obj):
        super().__init__()
        self.obj = obj

    def __str__(self):
        return (
            """Class '{obj.__module__}.{obj.__name__}' is not a plugin """
            """(it should inherit from """
            """'{superclass.__module__}.{superclass.__name__}')."""
            ).format(
                obj=self.obj,
                superclass=PluginBase,
            )

class PluginNotFound(errors.EvaristeError):
    """Plugin cannot be found."""

    def __init__(self, keyword):
        super().__init__()
        self.keyword = keyword

    def __str__(self):
        return "Cannot find plugin '{}'.".format(
            self.keyword,
            )

@functools.total_ordering
class PluginBase:
    """Plugin base: all imported plugins must be subclasses of this class."""
    # pylint: disable=too-few-public-methods

    keyword = None
    description = None
    priority = 0
    default_setup = {}
    global_default_setup = {}
    plugin_type = None
    subplugins = {}
    hooks = {}
    depends = {}
    default = False
    required = False

    def __init__(self, shared):
        self.hook_instances = {}
        self.shared = shared
        self.local = shared.get_plugin_view(self.keyword)
        self._set_default_setup()
        self._load_hooks()

    def _set_default_setup(self):
        """Set default value for this plugin setup, if necessary."""
        default = setup.Setup()
        for parent in reversed(self.__class__.mro()): # pylint: disable=no-member
            if hasattr(parent, "default_setup"):
                default.update(parent.global_default_setup)
                default.update({self.keyword: parent.default_setup})
        self.shared.setup.fill_blanks(default)

    def _load_hooks(self):
        """Load hooks."""
        for hook in self.hooks:
            self.hook_instances[hook] = functools.partial(self.hooks[hook], self)

    def match(self, value, *args, **kwargs): # pylint: disable=unused-argument
        """Return ``True`` iff ``value`` matches ``self``.

        Default is keyword match. This method can be overloaded by
        subclasses.
        """
        return value == self.keyword

    def __lt__(self, other):
        priority = self.priority
        if callable(priority):
            priority = priority() # pylint: disable=not-callable
        other_priority = other.priority
        if callable(other_priority):
            other_priority = other_priority()
        if priority == other_priority:
            return self.keyword < other.keyword
        return priority < other_priority

class PluginLoader:
    """Plugin loader. Find plugins and select the appropriate one."""

    def __init__(self, shared, libdirs=None):
        self.plugins = self._load_plugins(
            self._sort_plugins(
                self._list_available_plugins(libdirs)
                ),
            shared,
            )

    @staticmethod
    def _sort_plugins(pluginset):
        """Sort plugins by type and keyword."""
        plugindict = utils.DeepDict(2)

        for plugin in pluginset:
            if plugin.keyword in plugindict[plugin.plugin_type]:
                raise SameKeyword(
                    plugin.keyword,
                    plugin,
                    plugindict[plugin.plugin_type][plugin.keyword],
                    )

            plugindict[plugin.plugin_type][plugin.keyword] = plugin
        return plugindict

    @staticmethod
    def _load_plugins(available, shared):
        """Given the avaialble plugins, only loads relevant plugins.

        :return: A dictionary of loaded (instanciated) plugins.
        """
        # pylint: disable=too-many-branches, too-many-nested-blocks

        # Step 0: Converting setup options in a list of keyword
        for key in ["enable_plugins", "disable_plugins"]:
            if shared.setup['setup'][key] is None:
                shared.setup['setup'][key] = []
            else:
                if isinstance(shared.setup['setup'][key], str):
                    shared.setup['setup'][key] = shlex.split(shared.setup['setup'][key])
                elif isinstance(shared.setup['setup'][key], list):
                    pass
                else:
                    raise ValueError(
                        (
                            "'Setup[setup][enable_plugins]' should be a string or a "
                            "list (is now {}: '{}')."
                        ).format(
                            type(shared.setup['setup'][key]),
                            shared.setup['setup'][key],
                            ))

        # Step 1: Adding default plugins (unless disabled), and required plugins
        plugins = dict()
        for plugin_type in available:
            for keyword in available[plugin_type]:
                if (
                        available[plugin_type][keyword].default
                        and
                        keyword not in shared.setup['setup']['disable_plugins']
                    ):
                    plugins[keyword] = available[plugin_type][keyword]
                if available[plugin_type][keyword].required:
                    if keyword in shared.setup['setup']["disable_plugins"]:
                        LOGGER.warning(
                            (
                                "Disabling plugin '{}' is asked by setup, "
                                "but it is required. Still enabling."
                            ).format(keyword)
                            )
                    plugins[keyword] = available[plugin_type][keyword]

        # Step 2: Adding enabled plugins
        for keyword in shared.setup['setup']["enable_plugins"]:
            try:
                plugins[keyword] = available.get_subkey(keyword)
            except KeyError:
                raise PluginNotFound(keyword)

        # Step 3: Managing dependencies
        to_process = list(plugins.values())
        while to_process:
            plugin = to_process.pop()
            for keyword in plugin.depends.get('required', ()):
                try:
                    required = available.get_subkey(keyword)
                except KeyError:
                    raise PluginNotFound(keyword)
                if keyword not in plugins:
                    if keyword in shared.setup['setup']["disable_plugins"]: # pylint: disable=line-too-long
                        LOGGER.warning(
                            (
                                "Disabling plugin '{}' is asked by "
                                "setup, but it is required. Still "
                                "enabling."
                            ).format(keyword)
                        )
                    plugins[keyword] = required
                    to_process.append(required)
            for keyword in plugin.depends.get('suggested', ()):
                try:
                    suggested = available.get_subkey(keyword)
                except KeyError:
                    LOGGER.warning(
                        (
                            "Ignoring plugin '{}' required by "
                            "'{}': plugin not found."
                        ).format(
                            keyword, plugin.keyword,
                        ))
                    continue
                if suggested.keyword not in plugins:
                    if suggested.keyword in shared.setup['setup']["disable_plugins"]:
                        continue
                    plugins[suggested.keyword] = suggested
                    to_process.append(suggested)

        # Step 4: Instanciating plugins
        loaded = collections.defaultdict(dict)
        for keyword in plugins:
            loaded[plugins[keyword].plugin_type][keyword] = plugins[keyword](shared)

        return loaded

    @staticmethod
    def _list_available_plugins(libdirs):
        """Look for available plugins in accessible packages.

        Return a set of available plugins.
        """
        plugins = set()
        if libdirs is None:
            libdirs = []

        path = []
        path.extend([
            os.path.join(utils.expand_path(item), "plugins")
            for item
            in [".evariste", "~/.config/evariste", "~/.evariste"]
            ])
        path.extend([utils.expand_path(item) for item in libdirs])
        path.extend([
            os.path.join(item, "evariste")
            for item
            in sys.path
            ])

        for module in utils.iter_modules(path, "evariste.", LOGGER.debug):
            for attr in dir(module):
                if attr.startswith("_"):
                    continue
                obj = getattr(module, attr)

                if (
                        isinstance(obj, type)
                        and
                        issubclass(obj, PluginBase)
                    ):
                    if obj.keyword is None:
                        continue
                    plugins.add(obj)
        return plugins

    def match(self, plugin_type, value):
        """Return the first plugin matching ``value``.

        A plugin ``Foo`` matches ``value`` if ``Foo.match(value)`` returns
        True.
        """
        for plugin in sorted(self.get_plugin_type(plugin_type).values(), reverse=True):
            if plugin.match(value):
                return plugin
        raise NoMatch(value, sorted(self.iter_keywords(plugin_type)))

    def iter_pluginkeywords(self, plugin_type=None):
        """Iterate over plugin keywords."""
        for plugin in self.iter_plugins(plugin_type):
            yield plugin.keyword

    def iter_plugins(self, plugin_type=None):
        """Iterate over plugins."""
        if plugin_type is None:
            for ptype in self.plugins:
                for keyword in self.get_plugin_type(ptype):
                    yield self.get_plugin_type(ptype)[keyword]
        else:
            for keyword in self.get_plugin_type(plugin_type):
                yield self.get_plugin_type(plugin_type)[keyword]

    def iter_keywords(self, plugin_type=None):
        """Iterate over keywords"""
        if plugin_type is None:
            for ptype in self.plugins:
                yield from self.get_plugin_type(ptype)
        else:
            yield from self.get_plugin_type(plugin_type)

    def get_plugin_type(self, plugin_type):
        """Return a dictionary of plugins of the given type.

        This dictionary is indexed by plugin keywords.
        """
        return self.plugins[plugin_type]

    def get_plugin(self, keyword):
        """Return the plugin with the given keyword."""
        for plugin in self.iter_plugins():
            if plugin.keyword == keyword:
                return plugin
        raise KeyError(keyword)


def get_plugins(self, *_args, **_kwargs):
    """Given a :class:`AllPlugins` object, return itself.

    Useful to give hooks the plugins object.
    """
    return self

def get_libdirs(libdirs):
    """Convert `libdirs` setup option (as a string) to a list of path (as strings).
    """
    if libdirs:
        return shlex.split(libdirs)
    else:
        return []

class AllPlugins:
    """Gather all used plugins."""
    # pylint: disable=too-few-public-methods

    @set_methodhook("load_plugins", getter=get_plugins)
    def __init__(self, *, shared):
        self.shared = shared
        self.plugins = PluginLoader(
            shared,
            get_libdirs(shared.setup['setup']['libdirs']),
            )

    def iter_plugins(self, *args, **kwargs):
        """Iterate over loaded plugins."""
        yield from self.plugins.iter_plugins(*args, **kwargs)

    def _iter_hooks(self, name):
        """Iterator over functions registered for hook ``name``."""
        for plugin in self.iter_plugins():
            if name in plugin.hook_instances:
                yield plugin.hook_instances[name](self.shared)

    @contextlib.contextmanager
    def apply_context_hook(self, hookname, *, args=None, kwargs=None):
        """Apply a context hook (that is, run stuff before and after a `with` context).

        :param str hookname: Name of the hook to apply.
        :param list args: List of parameters to pass to hooks and
            ``function``.
        :param dict kwargs: Dictionary of parameters to pass to hooks and
            ``function``.
        """
        if args is None:
            args = list()
        if kwargs is None:
            kwargs = dict()
        hooks = None

        if hasattr(self, "plugins"):
            if hooks is None:
                hooks = list(self._iter_hooks(hookname))
            for hook in hooks:
                returnvalue = hook.enter(*args, **kwargs)
                if returnvalue is not None:
                    args, kwargs = returnvalue

        yield

        if hooks is None:
            hooks = list(self._iter_hooks(hookname))
        for hook in hooks:
            hook.exit()

    def apply_method_hook(self, hookname, function, *, args=None, kwargs=None):
        """Apply a method hook (that is, run stuff before and after a method call).

        :param str hookname: Name of the hook to apply.
        :param function function: Function to which the hook is to be applied.
        :param list args: List of parameters to pass to hooks and
            ``function``.
        :param dict kwargs: Dictionary of parameters to pass to hooks and
            ``function``.
        """
        if args is None:
            args = list()
        if kwargs is None:
            kwargs = dict()
        hooks = None

        if hasattr(self, "plugins"):
            if hooks is None:
                hooks = list(self._iter_hooks(hookname))
            for hook in hooks:
                returnvalue = hook.enter(*args, **kwargs)
                if returnvalue is not None:
                    args, kwargs = returnvalue

        returnvalue = function(*args, **kwargs)

        if hooks is None:
            hooks = list(self._iter_hooks(hookname))
        for hook in hooks:
            returnvalue = hook.exit(returnvalue)
        return returnvalue


    def get_plugin(self, keyword):
        """Return the plugin with the given keyword."""
        return self.plugins.get_plugin(keyword)

    def get_plugin_type(self, plugin_type):
        """Return a dictionary of plugins of the given type."""
        return self.plugins.get_plugin_type(plugin_type)

    def match_plugin(self, *args, **kwargs):
        """Return the first plugin matching the arguments.

        Arguments are transmitted to
        :meth:`evariste.plugins.PluginLoader.match`.
        """
        return self.plugins.match(*args, **kwargs)

    def iter_pluginkeywords(self):
        """Iterate over plugin keyword."""
        yield from self.plugins.iter_pluginkeywords()



# Copyright Louis Paternault 2015
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Test of evariste compilation"""

import logging
import os
import shutil

import evariste
from evariste.setup import Setup
from evariste.utils import ChangeDir

from . import TestCompilation

LOGGER = logging.getLogger(evariste.__name__)

DATA = os.path.join(
    os.path.dirname(__file__),
    'test_misc_data',
    )

class TestRun(TestCompilation):
    """Testing setup `libdirs` options"""
    # pylint: disable=too-few-public-methods

    def test_misc(self):
        """Test a miscallaenous compilation (nothing specific here)."""
        with ChangeDir(DATA):
            shutil.rmtree("dest", ignore_errors=True)
            os.makedirs("dest", exist_ok=True)
            self._run_evariste(Setup.from_file("evariste.setup"))
            try:
                self.assertTreeEqual(
                    "expected",
                    "dest",
                    )
            except AssertionError:
                LOGGER.error(
                    "Temporary tree has been kept in {}.".format(
                        os.path.join(DATA, "dest")
                        )
                    )
                raise



# Copyright Louis Paternault 2017
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Render tree as an HTML (body) page, with CSS"""

import os
import pkg_resources

from evariste import VERSION
from evariste import utils
from evariste.plugins.renderer.html import HTMLRenderer

class HTMLPlusRenderer(HTMLRenderer):
    """Render tree as an HTML div (without the `<div>` tags)."""
    # pylint: disable=too-few-public-methods

    keyword = "renderer.htmlplus"
    default_setup = {
        "template": "page.html",
        "display_log": "errors",
        }
    depends = {
        'required': [
            'copy',
        ],
        'suggested': [
            'renderer.html.readme.html',
            'renderer.html.target.image',
            'renderer.html.target.default',
        ],
    }
    default_templatevar = {
        "aftertree": """
            Generated using <a href="http://framagit.org/spalax/evariste">Évariste</a> version {}.
            Icons from the <a href="https://techbase.kde.org/Projects/Oxygen">KDE Oxygen</a> project.
            """.format(VERSION),
    }

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.environment.filters['labelize'] = self._labelize
        if 'copy_htmlplus' not in self.shared.setup['copy']:
            self.shared.setup['copy']['copy_htmlplus'] = (
                pkg_resources.resource_filename(
                    self.__class__.__module__,
                    os.path.join("data", "static"),
                    ),
                utils.expand_path(self.local.setup['staticdir']),
                )

    @staticmethod
    def _labelize(string):
        """Return a label, given a string.

        The label starts with alphabetical ascii characters, followed by
        digits. Two different strings have different hashes.
        """
        return "label" + str(hash(string))

    def _templatedirs(self):
        yield from super()._templatedirs()
        yield pkg_resources.resource_filename(
            "evariste.plugins.renderer.html",
            os.path.join("data", "templates"),
            )



# Copyright Louis Paternault 2015
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Test of the plugin dependencies"""

import os

from evariste.plugins import PluginNotFound

from . import TestLoadedPlugins
from .. import disable_logging

DEPENDSDIR = os.path.join(
    os.path.dirname(__file__),
    "depends",
    )

def string2set(string):
    """Turn a string into a set of stripped non-empty lines"""
    return set([
        line.strip() for line in string.split("\n") if line.strip()
        ])

class TestDefaultSetup(TestLoadedPlugins):
    """Testing default setup"""
    # pylint: disable=too-few-public-methods

    def test_noplugins(self):
        """Test default and required plugins."""
        with self.subTest():
            self.assertSetEqual(
                self._loaded_plugins({
                    'setup': {
                        'libdirs': DEPENDSDIR,
                        },
                    }),
                set([
                    'default',
                    'required',
                    ]) | self.default_plugins
                )

        with self.subTest():
            self.assertSetEqual(
                self._loaded_plugins({
                    'setup': {
                        'libdirs': DEPENDSDIR,
                        'disable_plugins': ['default', 'goodstuff'],
                        },
                    }),
                set([
                    'required',
                    ]) | self.default_plugins
                )

    def test_disable_required(self):
        """Exception is raised when disabling required plugin."""
        with self.assertLogs("evariste", level="WARNING") as log:
            self.assertSetEqual(
                self._loaded_plugins({
                    'setup': {
                        'libdirs': DEPENDSDIR,
                        'disable_plugins': ['required', 'goodstuff', 'default'],
                        },
                    }),
                set([
                    "required",
                    ]) | self.default_plugins,
                )
            self.assertEqual(
                log.output,
                [(
                    "WARNING:evariste:Disabling plugin 'required' is "
                    "asked by setup, but it is required. Still enabling."
                    )],
                )

    def test_non_existent_plugin(self):
        """Test exception when trying to enable a non-existent plugin."""
        with self.assertRaises(PluginNotFound):
            self._loaded_plugins({
                'setup': {
                    'libdirs': DEPENDSDIR,
                    'enable_plugins': ['nonexistent'],
                    },
                })

    def test_depends_simple(self):
        """Test simple dependencies"""
        with self.subTest():
            self.assertSetEqual(
                self._loaded_plugins({
                    'setup': {
                        'libdirs': DEPENDSDIR,
                        'disable_plugins': ['default', 'goodstuff'],
                        'enable_plugins': ['dependsoptional'],
                        },
                    }),
                set([
                    'required',
                    'dependsoptional',
                    'foo',
                    ]) | self.default_plugins
                )

        with self.subTest():
            self.assertSetEqual(
                self._loaded_plugins({
                    'setup': {
                        'libdirs': DEPENDSDIR,
                        'disable_plugins': ['default', 'goodstuff', 'foo'],
                        'enable_plugins': ['dependsoptional'],
                        },
                    }),
                set([
                    'required',
                    'dependsoptional',
                    ]) | self.default_plugins
                )

        with self.subTest():
            self.assertSetEqual(
                self._loaded_plugins({
                    'setup': {
                        'libdirs': DEPENDSDIR,
                        'disable_plugins': ['default', 'goodstuff'],
                        'enable_plugins': ['dependsrequired'],
                        },
                    }),
                set([
                    'dependsrequired',
                    'foo',
                    'required',
                    ]) | self.default_plugins
                )

        with disable_logging():
            with self.subTest():
                self.assertSetEqual(
                    self._loaded_plugins({
                        'setup': {
                            'libdirs': DEPENDSDIR,
                            'disable_plugins': ['default', 'goodstuff', 'foo'],
                            'enable_plugins': ['dependsrequired'],
                            },
                        }),
                    set([
                        'dependsrequired',
                        'foo',
                        'required',
                        ]) | self.default_plugins
                    )

    def test_depends_recursive(self):
        """Test recursive dependencies"""
        with disable_logging():
            with self.subTest():
                self.assertSetEqual(
                    self._loaded_plugins({
                        'setup': {
                            'libdirs': DEPENDSDIR,
                            'disable_plugins': ['default', 'goodstuff', 'foo'],
                            'enable_plugins': ['dependsrequiredrequired'],
                            },
                        }),
                    set([
                        'dependsrequiredrequired',
                        'dependsrequired',
                        'foo',
                        'required',
                        ]) | self.default_plugins
                    )

        with self.subTest():
            self.assertSetEqual(
                self._loaded_plugins({
                    'setup': {
                        'libdirs': DEPENDSDIR,
                        'disable_plugins': ['default', 'goodstuff'],
                        'enable_plugins': ['dependsrequiredoptional'],
                        },
                    }),
                set([
                    'dependsrequiredoptional',
                    'dependsoptional',
                    'foo',
                    'required',
                    ]) | self.default_plugins
                )

        with self.subTest():
            self.assertSetEqual(
                self._loaded_plugins({
                    'setup': {
                        'libdirs': DEPENDSDIR,
                        'disable_plugins': ['default', 'goodstuff', 'foo'],
                        'enable_plugins': ['dependsrequiredoptional'],
                        },
                    }),
                set([
                    'dependsrequiredoptional',
                    'dependsoptional',
                    'required',
                    ]) | self.default_plugins
                )

        with self.subTest():
            self.assertSetEqual(
                self._loaded_plugins({
                    'setup': {
                        'libdirs': DEPENDSDIR,
                        'disable_plugins': ['default', 'goodstuff'],
                        'enable_plugins': ['dependsoptionalrequired'],
                        },
                    }),
                set([
                    'dependsoptionalrequired',
                    'dependsrequired',
                    'foo',
                    'required',
                    ]) | self.default_plugins
                )

        with self.subTest():
            self.assertSetEqual(
                self._loaded_plugins({
                    'setup': {
                        'libdirs': DEPENDSDIR,
                        'disable_plugins': ['default', 'goodstuff', 'dependsrequired'],
                        'enable_plugins': ['dependsoptionalrequired'],
                        },
                    }),
                set([
                    'dependsoptionalrequired',
                    'required',
                    ]) | self.default_plugins
                )

        with self.subTest():
            self.assertSetEqual(
                self._loaded_plugins({
                    'setup': {
                        'libdirs': DEPENDSDIR,
                        'disable_plugins': ['default', 'goodstuff'],
                        'enable_plugins': ['dependsoptionaloptional'],
                        },
                    }),
                set([
                    'dependsoptional',
                    'dependsoptionaloptional',
                    'foo',
                    'required',
                    ]) | self.default_plugins
                )

        with self.subTest():
            self.assertSetEqual(
                self._loaded_plugins({
                    'setup': {
                        'libdirs': DEPENDSDIR,
                        'disable_plugins': ['default', 'goodstuff', 'foo'],
                        'enable_plugins': ['dependsoptionaloptional'],
                        },
                    }),
                set([
                    'dependsoptional',
                    'dependsoptionaloptional',
                    'required',
                    ]) | self.default_plugins
                )



# Copyright Louis Paternault 2017
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Shell command: perform a (list of) shell command(s) on files."""

import logging
import os
import pathlib
import re
import shutil
import shlex
import subprocess
import tempfile
import threading

from evariste import errors
from evariste.plugins.action import Action, PathCompiler, Report
from evariste.hooks import ContextHook
import evariste

LOGGER = logging.getLogger(evariste.__name__)
STRACE_RE = re.compile(r'^\d* *open\("(?P<name>.*)",.*O_RDONLY.*\) = *[^ -].*')

class MissingOption(errors.EvaristeError):
    """No command was provided for action :class:`Command`."""

    def __init__(self, filename, section, option):
        super().__init__()
        self.filename = filename
        self.section = section
        self.option = option

    def __str__(self):
        return (
            "Configuration for file '{file}' is missing option '{option}' in section '{section}'."
            ).format(
                file=self.filename,
                section=self.section,
                option=self.option,
            )

def system(command, path, log, depends, tempdir):
    """Run a system command.

    This function:
    - run command;
    - log standard output and error;
    - track opened files.
    """

    def _process_strace_line(line):
        """Process output line of strace, and complete ``depends`` if relevant."""
        match = STRACE_RE.match(line)
        if match:
            name = pathlib.Path(path.parent.from_fs) / pathlib.Path(match.groupdict()["name"])
            if name.resolve() != path.from_fs.resolve():
                if path.vcs.is_versionned(name):
                    depends.add(name)

    def _process_strace(pipe):
        """Process strace output, to find dependencies."""
        with open(pipe, mode="r", errors="replace") as file:
            for line in file:
                try:
                    _process_strace_line(line)
                except FileNotFoundError:
                    # File was deleted between the time its name is read and
                    # the time we check if it is a dependency: it can be
                    # discarded.
                    pass

    stdout = list(os.pipe())
    stderr = list(os.pipe())
    fifo = os.path.join(tempdir, str(id(path)))
    os.mkfifo(fifo)

    process = subprocess.Popen(
        r"""strace -f -o "{dest}" -e trace=open sh -c {command}""".format(
            dest=fifo,
            command=shlex.quote(command),
            ),
        shell=True,
        stdin=subprocess.DEVNULL,
        stdout=subprocess.PIPE,
        stderr=subprocess.STDOUT,
        pass_fds=stdout + stderr,
        # The following lines, as well as the `.decode(errors='replace')` a few
        # lines below), may be uncommented when support for python3.5 is
        # dropped.
        #universal_newlines=True,
        #errors="replace",
        cwd=path.parent.from_fs.as_posix(),
        )

    strace_thread = threading.Thread(
        target=_process_strace,
        daemon=True,
        kwargs={
            'pipe' : fifo,
            }
        )
    strace_thread.start()
    log.write(process.communicate()[0].decode(errors="replace"))
    os.unlink(fifo)
    strace_thread.join()
    for descriptor in stdout + stderr:
        os.close(descriptor)

    return process.returncode

class CommandCompiler(PathCompiler):
    """Run commands on the path."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._report = None # Temporary report. It will be completed during compilation.

    def iter_commands(self):
        """Iterator over the list of commands."""
        for option in self.path.config[self.keyword]:
            if option.startswith("command"):
                yield self.path.format(
                    self.path.config[self.keyword][option]
                    )

    @property
    def tempdir(self):
        """Return the path of a temporary, as secure as possible, directory."""
        return self.parent.tempdir

    def run_subcommand(self, command):
        """Run the subcommand ``command``."""
        LOGGER.info("Running command: {}".format(command))

        with self.shared.builder.lock.thread_safe():
            returncode = system(
                command=command,
                path=self._report.path,
                log=self._report.log,
                depends=self._report.depends,
                tempdir=self.tempdir,
                )
        return returncode == 0

    def get_target(self):
        """Return the path of the target."""
        if not self.path.config[self.keyword]['target'] is not None:
            raise MissingOption(self.path.from_fs, self.keyword, 'target')
        return (
            self.path.parent.from_source.as_posix()
            /
            pathlib.Path(self.path.format(self.path.config[self.keyword]["target"]))
            )

    def compile(self):
        """Perform compilation of `self.path`."""
        self._report = Report(
            self.path,
            success=True,
            target=self.get_target(),
            )
        for command in self.iter_commands():
            parsed_command = self.path.format(command)
            self._report.log.write("$ {}\n".format(parsed_command))
            if not self.run_subcommand(parsed_command):
                self._report.success = False
                break
        return self._report

class TempDir(ContextHook):
    """Manage creation and deletion of a temporary directory."""

    def enter(self, *__args, **__kwargs):
        self.plugin.tempdir = tempfile.mkdtemp(
            prefix="evariste-{}-{}-".format(os.getpid(), self.plugin.keyword),
            )

    def exit(self, value=None):
        shutil.rmtree(self.plugin.tempdir)
        return value

class Command(Action):
    """Chain of commands"""

    keyword = "action.command"
    pathcompiler = CommandCompiler
    hooks = {
        'compiletree': TempDir,
        }

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.tempdir = None

    def match(self, value):
        return False



#!/usr/bin/env python3

# Copyright 2015 Louis Paternault
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Installer"""

from setuptools import setup, find_packages
import codecs
import glob
import os
import sys

def readme():
    directory = os.path.dirname(os.path.join(
        os.getcwd(),
        __file__,
        ))
    with codecs.open(
        os.path.join(directory, "README.rst"),
        encoding="utf8",
        mode="r",
        errors="replace",
        ) as file:
        return file.read()

setup(
        name='Evariste',
        version="0.2.1",
        packages=find_packages(exclude=["test*"]),
        setup_requires=["hgtools"],
        install_requires=[
            "argdispatch",
            "jinja2",
            "docutils",
            "gitpython",
            "markdown",
            "unidecode",
            ],
        include_package_data=True,
        author='Louis Paternault',
        author_email='spalax+python@gresille.org',
        description='TODO',
        url='https://git.framasoft.org/spalax/evariste',
        license="AGPLv3 or any later version",
        test_suite="test.suite",
        entry_points={
            'console_scripts':
                [
                    'evariste = evariste.evs.compile.__main__:main',
                    'evs = evariste.evs.__main__:main',
                ]
                },
        classifiers=[ # TODO
            "Development Status :: 3 - Alpha",
            "Programming Language :: Python",
            "Programming Language :: Python :: 3",
            "Programming Language :: Python :: 3.4",
            "Programming Language :: Python :: 3.5",
            "Programming Language :: Python :: 3.6",
            "License :: OSI Approved :: GNU Affero General Public License v3 or later (AGPLv3+)",
        ],
        long_description=readme(),
        zip_safe = False,
)



# Copyright Louis Paternault 2016
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Directory representation and compilation."""

import configparser
import functools
import logging
import os
import pathlib
import queue
import tarfile

from evariste import plugins, utils, errors, multithread
from evariste.plugins.action import DirectoryAction
from evariste.hooks import set_methodhook

LOGGER = logging.getLogger(__name__)

def get_tree_plugins(self, *args, **kwargs):
    """Given the argument of :meth:`Tree.__init__`, return a :class:`AllPlugins` instance."""
    # pylint: disable=unused-argument
    if isinstance(self, Root):
        root = self
    else:
        root = kwargs['parent']
        while root.parent is not None:
            root = root.parent
    return root.shared.builder.plugins

@functools.total_ordering
class Tree:
    """Directory tree"""
    # pylint: disable=too-many-instance-attributes

    @set_methodhook("buildtreeitem", getter=get_tree_plugins)
    def __init__(self, path, *, parent):
        self._subpath = {}
        self.report = None
        self.ignored = []
        self.parent = parent
        self.config = None
        if parent is not None:
            self.basename = pathlib.Path(path)
            self.from_fs = parent.from_fs / self.basename
            self.from_source = parent.from_source / self.basename
            self.vcs = self.parent.vcs
            self.shared = self.parent.shared
        self.local = self.shared.get_tree_view(path)
        self._set_ignore()
        self._set_config()

    def __hash__(self):
        return hash((hash(self.parent), self.from_fs))

    @property
    def relativename(self):
        """Return a 'relative' path.

        * For root, return path relative to file system (or directory of setup file).
        * For non-root, return path relative to parent (i.e. basename of path).
        """
        if isinstance(self, Root):
            return self.from_fs
        else:
            return self.basename

    @staticmethod
    def is_root():
        """Return ``True`` iff ``self`` is the root."""
        return False

    @property
    def depth(self):
        """Return the depth of the path.

        The root has depth 0, and depth of each path is one more than the depth
        of its parent.
        """
        if isinstance(self, Root):
            return 0
        else:
            return 1 + self.parent.depth

    @property
    def root(self):
        """Return the root of the tree."""
        if self.is_root():
            return self
        return self.parent.root

    def find(self, path):
        """Return ``True`` iff ``path`` exists, as a subpath of ``self``.

        Argument can be:
        - a string (:type:`str`);
        - a :type:`pathlib.Path` object;
        - a tuple of strings, as a list of directories and (optional) final file.
        """
        if isinstance(path, str):
            pathtuple = pathlib.Path(path)
        elif isinstance(path, pathlib.Path):
            pathtuple = path.parts
        else:
            pathtuple = path

        if not pathtuple:
            return True
        if pathtuple[0] in self:
            return self[pathtuple[0]].find(pathtuple[1:])
        return False

    def add_subpath(self, sub):
        """Add a subpath"""
        if len(sub.parts) > 1:
            self[sub.parts[0]].add_subpath(pathlib.Path(*sub.parts[1:]))
        else:
            self[sub.parts[0]] # pylint: disable=pointless-statement

    def __iter__(self):
        return iter(self._subpath)

    def keys(self):
        """Iterator over subpaths (as :class:`str` objects)."""
        yield from self._subpath.keys()

    def values(self):
        """Iterator over subpaths (as :class:`Tree` objects)."""
        yield from self._subpath.values()

    def __contains__(self, key):
        return key in self._subpath

    def __getitem__(self, key):
        if key not in self:
            if (self.from_fs / key).is_dir():
                path_type = Directory
            else:
                path_type = File
            self._subpath[key] = path_type(key, parent=self)
        return self._subpath[key]

    def __delitem__(self, item):
        pathitem = pathlib.Path(item)
        if len(pathitem.parts) != 1:
            raise errors.EvaristeBug(
                "Argument '{}' should be a single directory or file, "
                "not a 'directory/directory' or directory/file'."
                )
        if str(pathitem) in self:
            del self._subpath[str(pathitem)]
            if (not self._subpath) and self.is_dir() and self.parent is not None:
                del self.parent[self.basename]


    def __str__(self):
        return self.from_fs.as_posix()

    def __len__(self):
        return len(self._subpath)

    def __eq__(self, other):
        return self.from_source == other.from_source

    def __lt__(self, other):
        if isinstance(self, Directory) and not isinstance(other, Directory):
            return True
        if not isinstance(self, Directory) and isinstance(other, Directory):
            return False
        return self.from_source < other.from_source

    def is_dir(self):
        """Return `True` iff `self` is a directory."""
        return issubclass(self.__class__, Directory)

    def is_file(self):
        """Return `True` iff `self` is a file."""
        return issubclass(self.__class__, File)

    def walk(self, dirs=False, files=True):
        """Iterator over files or directories  of `self`.

        :param bool dirs: If `False`, do not yield directories.
        :param bool files: If `False`, do not yield files.
        """
        if (
                (dirs and self.is_dir())
                or
                (files and self.is_file())
            ):
            yield self
        for sub in sorted(self):
            yield from self[sub].walk(dirs, files)

    def iter_ignored(self):
        """Iterate over ignored paths."""
        # Recursive call
        for sub in self:
            yield from self[sub].iter_ignored()

        # Process ``self``
        for regexp in self.ignored:
            for file in self.from_fs.glob(str(regexp)):
                yield file.relative_to(self.root.from_fs)

        # Process ``self``
        for file in self.config_files(parent=False):
            yield pathlib.Path(file).relative_to(self.root.from_fs)

    def prune(self, path):
        """Remove a file.

        Argument can be either a :class:`pathlib.Path` or a :class:`tuple`.
        """
        if isinstance(path, tuple):
            parts = path
        elif isinstance(path, pathlib.Path):
            parts = path.parts
        else:
            raise ValueError
        if len(parts) == 1:
            del self[parts[0]]
        else:
            self[parts[0]].prune(parts[1:])

    def _parent_config_files(self):
        """Iterate over the names of configuration files affecting parents of ``self``.
        """
        if self.parent is not None:
            yield from self.parent.config_files(parent=True)

    def _set_ignore(self):
        """Set the list of ignored regular expressions."""
        raise NotImplementedError()

    def _set_config(self):
        """Set the configuration of file or directory."""
        # Set configuration of `self`
        config = configparser.ConfigParser(allow_no_value=True)
        for filename in self.config_files(parent=True):
            with open(filename) as file:
                config.read_file(file)
        self.config = utils.DeepDict.from_configparser(config)


    def config_files(self, parent=False):
        """Iterator over the list of configuration files."""
        raise NotImplementedError()

    def iter_depends(self):
        """Iterate files that are dependencies of others."""
        raise NotImplementedError()

    def format(self, string):
        """Format given string, with several variables related to ``self``.
        """
        suffix = self.from_source.suffix
        if suffix.startswith("."):
            suffix = suffix[1:]
        return string.format(
            dirname=self.parent.from_fs.as_posix(),
            filename=self.basename,
            fullname=self.from_fs.as_posix(),
            extension=suffix,
            basename=self.basename.stem,
            )


    def render_readme(self, renderers, *args, **kwargs):
        """Find and render the readme corresponding to ``self``.

        If no readme was found, return ``None``.
        """
        for renderer in renderers:
            for filename in renderer.iter_readme(
                    self.from_fs,
                    directory=self.is_dir(),
                ):
                return renderer.render(filename, *args, **kwargs)
        return None

    def iter_readmes(self, renderers):
        """Iterate over readme files."""
        for renderer in renderers:
            for filename in renderer.iter_readme(
                    self.from_fs,
                    directory=self.is_dir(),
                ):
                yield pathlib.Path(filename).relative_to(self.root.from_fs)
                return

    def full_depends(self):
        """Return the list of all dependencies of this tree (recursively for directories)."""
        raise NotImplementedError()


class Directory(Tree):
    """Directory"""

    def _set_ignore(self):
        """Set the list of ignored regular expressions."""
        ignorename = os.path.join(self.from_fs.as_posix(), ".evsignore")
        if os.path.exists(ignorename):
            self.ignored.append(os.path.basename(ignorename))
            with open(ignorename) as file:
                for line in file.readlines():
                    line = line.strip()
                    if not line:
                        continue
                    if line.startswith("#"):
                        continue
                    try:
                        # Fake glob, to catch lines with invalid syntax.
                        for _ in self.from_fs.glob(line):
                            break
                    except ValueError as error:
                        LOGGER.warning("Line '{}' has been ignored in file '{}': {}.".format(
                            line,
                            ignorename,
                            str(error),
                            ))
                        continue
                    self.ignored.append(line)

    def config_files(self, parent=False):
        """Iterator over the names of configuration files specific to ``self``
        """
        if parent:
            yield from self._parent_config_files()

        name = os.path.join(self.from_fs.as_posix(), ".evsconfig")
        if os.path.exists(name):
            yield name

    def feed_compile_tasks(self, *, taskqueue, to_prune, to_report):
        """Put tasks corresponding to actions to perform on files, in the queue.

        :param multithread.Queue taskqueue: Queue of tasks to perform in
            parallel, now.
        :param multithread.Queue to_prune: Queue of tasks to perform when
            every task of ``taskqueue`` has been executed, to prune files that
            are to be ignored.
        :param multithread.Queue to_report: Queue of tasks to perform when
            every task of ``taskqueue`` has been executed, to build the
            :attr:`Tree.report` object.
        """
        priority = 0
        for sub in sorted(self):
            if self[sub].is_file():
                taskqueue.put(
                    functools.partial(self[sub].compile, to_prune=to_prune)
                    )
            else:
                priority = max(
                    priority,
                    self[sub].feed_compile_tasks(
                        taskqueue=taskqueue,
                        to_prune=to_prune,
                        to_report=to_report,
                        )
                    )

        to_report.put(multithread.Task(
            (1, priority+1),
            self._set_report,
            ))

        return priority + 1

    def _set_report(self):
        """Build and set `self.report`."""
        self.report = DirectoryAction(self.shared).compile(self)

    def iter_depends(self):
        """Iterate files that are dependencies of others."""
        # We cannot use `for sub in self:` here because dictionnary
        # `self._subpath` may change during iteration.
        for sub in list(self):
            if sub in self:
                yield from self[sub].iter_depends()

    def iter_readmes(self, renderers):
        renderers = list(renderers)
        yield from super().iter_readmes(renderers)

        for sub in self:
            yield from self[sub].iter_readmes(renderers)

    def full_depends(self):
        for sub in self:
            yield from self[sub].full_depends()


class File(Tree):
    """File"""

    def _set_ignore(self):
        """Set the list of ignored regular expressions."""
        for name in ["{}.ignore", ".{}.ignore"]:
            if os.path.exists(os.path.join(
                    self.parent.from_fs.as_posix(),
                    name.format(self.basename),
                )):
                self.parent.ignored.append(self.basename)
                self.parent.ignored.append(name.format(self.basename))

    def config_files(self, parent=False):
        if parent:
            yield from self._parent_config_files()

        for name in [".{}.evsconfig", "{}.evsconfig"]:
            name = os.path.join(
                self.parent.from_fs.as_posix(),
                name.format(self.basename),
                )
            if os.path.exists(name):
                yield name

    def compile(self, *, to_prune):
        """Compile file."""
        with self.shared.builder.lock.thread_unsafe():
            LOGGER.info("Compiling '{}'…".format(self.from_source))
            try:
                if not self.shared.builder.plugins.get_plugin('changed').compile(self):
                    # Nothing has changed: it is useless to compile this again.
                    compiler = self.shared.builder.plugins.get_plugin('action.useless')
                elif self.config['action']['plugin'] is not None:
                    compiler = self.shared.builder.plugins.get_plugin('{}.{}'.format(
                        'action',
                        self.config['action']['plugin'],
                        ))
                else:
                    compiler = self.shared.builder.plugins.match_plugin('action', self.from_fs)
            except plugins.NoMatch:
                to_prune.put(multithread.Task(
                    (0, -self.depth),
                    functools.partial(
                        self.parent.__delitem__,
                        self.basename,
                    )
                    ))
                LOGGER.info("Compiling '{}': ignored.".format(self.from_source))
                return

            with self.shared.builder.plugins.apply_context_hook("compilefile", args=[self]):
                self.report = compiler.compile(self)
            if self.report.success:
                LOGGER.info("Compiling '{}': success.".format(self.from_source))
            else:
                LOGGER.info("Compiling '{}': failed.".format(self.from_source))

    def iter_depends(self):
        """Iterate over dependencies."""
        if self.report.success:
            for path in self.report.depends:
                yield path.resolve().relative_to(
                    self.root.from_fs.resolve()
                    )

    @set_methodhook("makearchive")
    @functools.lru_cache(maxsize=None)
    def make_archive(self, destdir):
        """Make an archive of ``self`` and its dependency.

        Stes are:
        - build the archive;
        - copy it to ``destdir``;
        - return the path of the archive, relative to ``destdir``.

        If ``self`` has no dependencies, consider the file as an archive.

        It can be called several times: the archive will be built only once.
        """

        def common_root(files):
            """Look for the common root of files given in argument.

            Returns a tuple of `(root, relative_files)`, where
            `relative_files` is a list of `files`, relative to the root.
            """
            files = [
                file.resolve()
                for file
                in files
                if file.resolve().as_posix().startswith(self.root.from_fs.resolve().as_posix())
                ]
            root = pathlib.Path()
            while True:
                prefixes = [path.parts[0] for path in files]
                if len(set(prefixes)) != 1:
                    break
                prefix = prefixes[0]
                files = [
                    file.relative_to(prefix)
                    for file
                    in files
                    ]
                root /= prefix
            return root.relative_to(self.root.from_fs.resolve()), files

        if (
                # Compilation was successful
                self.report.success
                # File has not been compiled again
                and not self.shared.builder.plugins.get_plugin('changed').compile(self)
                # Archive has already been generated
                and 'archivepath' in self.shared.tree[self.from_source]['changed']
            ):
            return self.shared.tree[self.from_source]['changed']['archivepath']

        if len(self.report.full_depends) == 1:
            utils.copy(self.from_fs.as_posix(), (destdir / self.from_source).as_posix())
            return self.from_source
        else:
            archivepath = self.from_source.with_suffix(
                "{}.{}".format(self.from_source.suffix, 'tar.gz')
                )
            os.makedirs(os.path.dirname((destdir / archivepath).as_posix()), exist_ok=True)

            archive_root, full_depends = common_root(self.report.full_depends)
            with tarfile.open(
                (destdir / archivepath).as_posix(),
                mode='w:gz',
                ) as archive:
                for file in full_depends:
                    archive.add(
                        (self.root.from_fs / archive_root / file).as_posix(),
                        file.as_posix(),
                        )
            return archivepath

    def last_modified(self):
        """Return the last modified date and time of ``self``."""
        return self.vcs.last_modified(self.from_fs)

    def full_depends(self):
        yield from self.report.full_depends

    def depends(self):
        """Iterator over dependencies of this file (but not the file itself)."""
        yield from self.report.depends

class Root(Directory):
    """Root object (directory with no parents)"""

    def __init__(self, path, *, vcs=None, shared=None):
        self.vcs = vcs
        self.shared = shared
        self.from_fs = path
        self.from_source = pathlib.Path(".")
        self.basename = pathlib.Path(".")
        super().__init__(path, parent=None)

    @staticmethod
    def is_root():
        return True

    def root_compile(self):
        """Recursively compile files.."""
        self.prune_ignored()
        workers = self.shared.setup["arguments"]["jobs"]
        with multithread.Pool(workers) as pool:
            to_prune = queue.PriorityQueue()
            to_report = queue.PriorityQueue()
            self.feed_compile_tasks(
                taskqueue=pool.queue,
                to_prune=to_prune,
                to_report=to_report
                )
        while not to_prune.empty():
            to_prune.get().run()
        self.prune_depends()
        while not to_report.empty():
            to_report.get().run()

    def prune_ignored(self):
        """Prune ignored files."""
        for file in list(self.iter_ignored()):
            if self.find(file):
                self.prune(file)

    def prune_depends(self):
        """Prune dependencies."""
        for file in list(self.iter_depends()):
            if self.find(file):
                self.prune(file)

    @classmethod
    def from_vcs(cls, repository):
        """Return a directory, fully set."""
        tree = cls(repository.source, vcs=repository, shared=repository.shared)
        for path in repository.walk():
            tree.add_subpath(pathlib.Path(path))
        return tree



# Copyright Louis Paternault 2015
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Test of the setup plugin"""

import unittest

from evariste.plugins import PluginBase
from evariste.setup import Setup
from evariste.shared import Shared

################################################################################

class First(PluginBase):
    """Test plugin"""
    # pylint: disable=abstract-method, too-few-public-methods

    plugin_type = 'test'
    default_setup = {
        'foo': 1,
        'bar': 2,
        }

class Second(First):
    """Another test plugin"""
    # pylint: disable=abstract-method, too-few-public-methods

    keyword = "test.second"
    default_setup = {
        'foo': 3,
        'baz': 4,
        }

class TestDefaultSetup(unittest.TestCase):
    """Testing default setup"""
    # pylint: disable=too-few-public-methods

    def test_inheritance(self):
        """Testing default setup inheritance"""
        plugin = Second(Shared(
            setup=Setup.from_dict({'setup': {'vcs': 'vcs.none'}}),
            builder=None,
            ))
        self.assertEqual(plugin.shared.setup['test.second']['foo'], 3)
        self.assertEqual(plugin.shared.setup['test.second']['bar'], 2)
        self.assertEqual(plugin.shared.setup['test.second']['baz'], 4)

        self.assertEqual(plugin.local.setup['foo'], 3)
        self.assertEqual(plugin.local.setup['bar'], 2)
        self.assertEqual(plugin.local.setup['baz'], 4)

################################################################################

class TestEqual(unittest.TestCase):
    """Test equality method of setup objects."""

    def test_empty(self):
        """Test the :meth:`Setup.__eq__` method on empty setup objects."""
        self.assertEqual(
            Setup.from_dict({}),
            Setup.from_dict({}),
            )

    def test_same_dict(self):
        """Test the :meth:`Setup.__eq__` method on object built with the same dictionary."""
        base = {
            "first": {
                "one": 11,
                "two": 12,
                },
            "second": {
                "one": 21,
                "two": 22,
                }
            }
        self.assertEqual(
            Setup.from_dict(base),
            Setup.from_dict(base),
            )

    def test_default_objects(self):
        """Test the :meth:`Setup.__eq__` method on object having default values."""
        setup1 = Setup.from_dict({
            "first": {
                "one": 11,
                "two": 12,
                },
            "second": {
                "one": 21,
                "two": 22,
                }
            })
        setup2 = Setup.from_dict(vars(setup1))

        # Creating default objects
        setup1["three"] # pylint: disable=pointless-statement
        setup1["one"]["three"] # pylint: disable=pointless-statement
        setup1["four"]["one"] # pylint: disable=pointless-statement

        self.assertEqual(setup1, setup2)

    def test_different(self):
        """Test the :meth:`Setup.__eq__` method on different objects."""
        setup1 = Setup.from_dict({
            "first": {
                "one": 11,
                "two": 12,
                },
            "second": {
                "one": 21,
                "two": 22,
                }
            })

        with self.subTest():
            setup2 = setup1.copy()
            setup2["first"]["one"] = "foo"
            self.assertNotEqual(setup1, setup2)

        with self.subTest():
            setup3 = setup1.copy()
            setup3["first"]["three"] = "foo"
            self.assertNotEqual(setup1, setup3)

        with self.subTest():
            setup4 = setup1.copy()
            setup4["three"]["one"] = 31
            self.assertNotEqual(setup1, setup4)


################################################################################

class TestUpdate(unittest.TestCase):
    """Testing method :meth:`setup.Setup.update`."""

    def test_update(self):
        """Test the :meth:`Setup.update` method."""
        setup = Setup.from_dict({
            "first": {
                "one": 11,
                "two": 12,
                },
            "second": {
                "one": 21,
                "two": 22,
                }
            })
        setup.update({
            "first": {
                "one": "UPDATED",
                "three": 13,
                },
            "three": {
                "one": 31,
                },
            })
        self.assertEqual(
            setup,
            Setup.from_dict({
                "first": {
                    "one": "UPDATED",
                    "two": 12,
                    "three": 13,
                    },
                "second": {
                    "one": 21,
                    "two": 22,
                    },
                "three": {
                    "one": 31,
                    },
                })
            )
